

 /*
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */ 
 
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc. All rights reserved.
 *
 */
 
/*
 * Module Name:    mfw_gst_streaming_cache.c
 *
 * Description:    Implementation for streamed based demuxer srcpad cache.
 *
 * Portability:    This code is written for Linux OS.
 */  
 
/*
 * Changelog: 
 *
 */





#include <gst/gst.h>
#include <gst/base/gstadapter.h>

#include "aiurdemux.h"



#define WAIT_COND_TIMEOUT(cond, mutex, timeout) \
    do{\
        GTimeVal now;\
        g_get_current_time(&now);\
        g_time_val_add(&now, (glong)(timeout));\
        g_cond_timed_wait((cond),(mutex),&now);\
    }while(0)

static GstAiurStreamCacheClass * aiur_stream_cache_parent_class = NULL;

static GTimeVal timeout = {1,0};

GType gst_aiur_stream_cache_get_type(void);

void
gst_aiur_stream_cache_finalize(GstAiurStreamCache * cache)
{
    if (cache->pad){
        gst_object_unref(GST_OBJECT_CAST(cache->pad));
    }
    gst_adapter_clear(cache->adapter);
    gst_object_unref(cache->adapter);
    g_cond_free(cache->produce_cond);
    g_cond_free(cache->consume_cond);
    g_mutex_free(cache->mutex);
}

void
gst_aiur_stream_cache_close(GstAiurStreamCache * cache)
{
    if (cache){
        cache->close = TRUE;
    }
}

void
gst_aiur_stream_cache_open(GstAiurStreamCache * cache)
{
    if (cache){
        cache->close = FALSE;
    }
}



GstAiurStreamCache * 
gst_aiur_stream_cache_new(guint64 maxsize)
{
    GstAiurStreamCache * cache  = gst_mini_object_new(GST_TYPE_AIURSTREAMCACHE);

    cache->pad = NULL;
    
    cache->adapter=gst_adapter_new();
    cache->mutex = g_mutex_new();
    cache->consume_cond = g_cond_new();
    cache->produce_cond = g_cond_new();
    
    cache->maxsize = maxsize;
    cache->offset = 0;
    cache->size = 0;
    cache->eos = FALSE;
    cache->seeking = FALSE;
    cache->close = FALSE;
    cache->drop = 0;
    cache->length = 0;
    
    return cache;
}

void 
gst_aiur_stream_cache_attach_pad(GstAiurStreamCache * cache, GstPad * pad)
{
    if (cache){
        g_mutex_lock(cache->mutex);
        if (cache->pad){
            gst_object_unref(GST_OBJECT_CAST(cache->pad));
        }

        if (pad){
            gst_object_ref(GST_OBJECT_CAST(pad));

            GstQuery *q;
            GstFormat fmt;

            q = gst_query_new_seeking (GST_FORMAT_BYTES);
            if (gst_pad_peer_query (pad, q)) {
                gst_query_parse_seeking (q, &fmt, &cache->seekable, NULL,
                                         NULL);
            }
            gst_query_unref (q);

            q = gst_query_new_duration(GST_FORMAT_BYTES);
            if (gst_pad_peer_query (pad, q)) {
                gint64 duration;
                gst_query_parse_duration(q, &fmt,&duration);
                cache->length = duration;
            }
            gst_query_unref (q);
        }

        cache->pad = pad;
        
        g_mutex_unlock(cache->mutex);
    }
}

gint64 
gst_aiur_stream_cache_availiable_bytes(GstAiurStreamCache * cache)
{
    gint64 avail = -1;

    
    if (cache){
        g_mutex_lock(cache->mutex);
        avail = cache->size;
        g_mutex_unlock(cache->mutex);
    }
    
    return avail;
}

void 
gst_aiur_stream_cache_set_segment(GstAiurStreamCache * cache, guint64 start, guint64 stop)
{
    
    if (cache){
        
        g_mutex_lock(cache->mutex);
        
        cache->seeking = FALSE;
        cache->offset = start;
        gst_adapter_clear(cache->adapter);
        cache->size = 0;
        cache->eos=FALSE;
        
        g_cond_signal(cache->consume_cond);
        
        g_mutex_unlock(cache->mutex);
    }
}

void 
gst_aiur_stream_cache_get_status(GstAiurStreamCache * cache, gint32 * size, gint32 * maxsize)
{
    if (cache){
        
        g_mutex_lock(cache->mutex);
        
        * size = (gint32)cache->size;
        * maxsize = (gint32)cache->maxsize;
        
        g_mutex_unlock(cache->mutex);
    }
}


void 
gst_aiur_stream_cache_add_buffer(GstAiurStreamCache * cache, GstBuffer * buffer)
{
    guint64 size;
    if ((cache==NULL)||(buffer==NULL))
        goto bail;
    g_mutex_lock(cache->mutex);

    size=GST_BUFFER_SIZE(buffer);

    
    if ((cache->seeking) || (size==0)){
        g_mutex_unlock(cache->mutex);
        goto bail;
    }
    if (cache->drop){
        if (cache->drop>=size){
            cache->drop-=size;
            g_mutex_unlock(cache->mutex);
            goto bail;
        }else{
            GST_BUFFER_DATA(buffer)+=(size-cache->drop);
            GST_BUFFER_SIZE(buffer)-=(size-cache->drop);
            size = GST_BUFFER_SIZE(buffer);
            cache->drop = 0;
        }
    }
    if (cache->maxsize){
        if (cache->maxsize<size){
            cache->maxsize = size;
        }
    
        while ((cache->size>cache->maxsize)&&(cache->close==FALSE)){
            WAIT_COND_TIMEOUT(cache->consume_cond, cache->mutex, 1000000);
        }

        if (cache->seeking){
            g_mutex_unlock(cache->mutex);
            goto bail;
        }
    }


    cache->size+=size;

    gst_adapter_push(cache->adapter, buffer);

    
    g_cond_signal(cache->produce_cond);
    
    g_mutex_unlock(cache->mutex);
    
    return;    


bail:
    if (buffer){
        gst_buffer_unref(buffer);
    }
}

void 
gst_aiur_stream_cache_seteos(GstAiurStreamCache * cache, gboolean eos)
{
    
    if (cache){
        
        g_mutex_lock(cache->mutex);
        
        cache->eos = eos;
        g_cond_signal(cache->produce_cond);
        
        g_mutex_unlock(cache->mutex);
    }
}

gint64
gst_aiur_stream_cache_get_position(GstAiurStreamCache * cache)
{
    gint64 pos = -1;
    if (cache){
        
        g_mutex_lock(cache->mutex);
        pos = cache->offset;
        g_mutex_unlock(cache->mutex);
    }
    return pos;
}


gint
gst_aiur_stream_cache_seek(GstAiurStreamCache * cache, guint64 addr)
{
    gint r = 0;
    gboolean ret;
    guint flushsize;
    
    if (cache==NULL){
        return -1;
    }
    

    g_mutex_lock(cache->mutex);
    
    if (addr==cache->offset){
        g_mutex_unlock(cache->mutex);
        return r;
    }


    gst_adapter_clear(cache->adapter);
    g_cond_signal(cache->consume_cond);
    
    cache->size = 0;
    cache->offset = addr;

    cache->seeking=TRUE;
    cache->eos=FALSE;
    g_mutex_unlock(cache->mutex);
    ret =  gst_pad_push_event (cache->pad, gst_event_new_seek ((gdouble)1, GST_FORMAT_BYTES, GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET, (gint64)addr,
      GST_SEEK_TYPE_NONE, (gint64)(-1)));

    if (ret==FALSE){
        r = -1;
    }
    return r;
}

#define READ_BYTES(cache, buffer, readbytes)\
    do {\
        gst_adapter_copy((cache)->adapter, (buffer), 0, (readbytes));\
        gst_adapter_flush((cache)->adapter, (readbytes));\
        (cache)->size-=(readbytes);\
        (cache)->offset+=(readbytes);\
        g_cond_signal((cache)->consume_cond);\
    }while(0)


gint64 
gst_aiur_stream_cache_read(GstAiurStreamCache * cache, guint64 size, char * buffer)
{
    gint64 readsize = -1;
    
    if (cache==NULL){
        return readsize;
    }


try_read:

    if (cache->close==TRUE){
        return readsize;
    }
    
    g_mutex_lock(cache->mutex);

    if (cache->seeking==TRUE)
        goto not_enough_bytes;

    if ((cache->maxsize) && (cache->maxsize<size)){
        cache->maxsize = size;
        /* enlarge maxsize means consumed */
        g_cond_signal(cache->consume_cond);
    }
    
    if (size>cache->size){
        if (cache->eos){/* not enough bytes when eos */
            readsize = cache->size;
            READ_BYTES(cache, buffer, readsize);
            goto beach;
        }
        goto not_enough_bytes;
    }

    readsize = size;
    READ_BYTES(cache, buffer, readsize);

    goto beach;


not_enough_bytes:
    //g_print("not enough\n");
    WAIT_COND_TIMEOUT(cache->produce_cond, cache->mutex, 1000000);
    g_mutex_unlock(cache->mutex);

    goto try_read;


beach:
    g_mutex_unlock(cache->mutex);
    return readsize;
}



static void
gst_aiur_stream_cache_class_init (gpointer g_class, gpointer class_data)
{
  GstMiniObjectClass *mini_object_class = GST_MINI_OBJECT_CLASS (g_class);

  aiur_stream_cache_parent_class = g_type_class_peek_parent (g_class);

  mini_object_class->finalize = (GstMiniObjectFinalizeFunction)
      gst_aiur_stream_cache_finalize;
}


GType gst_aiur_stream_cache_get_type(void)
{

    static GType aiur_stream_cache_type = 0;
    if (G_UNLIKELY(aiur_stream_cache_type == 0)) {
    static const GTypeInfo aiurdemux_stream_cache_info = {
        sizeof(GstAiurStreamCacheClass),
        NULL,
        NULL,
        gst_aiur_stream_cache_class_init,
        NULL,
        NULL,
        sizeof(GstAiurStreamCache),
        0,
        NULL,
        NULL
    };
    aiur_stream_cache_type =
        g_type_register_static(GST_TYPE_MINI_OBJECT, "AiurStreamCache",
                   &aiurdemux_stream_cache_info, 0);
    }

    return aiur_stream_cache_type;
}




